/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

#ifndef _RENDERCHECK_GL_H_
#define _RENDERCHECK_GL_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <vector>
#include <map>
#include <string>

#include <GL/glew.h>

#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <nvShaderUtils.h>

using std::vector;
using std::map;
using std::string;

#define BUFFER_OFFSET(i) ((char *)NULL + (i))


class CheckRender
{
public:
    CheckRender() {}

	CheckRender(unsigned int width, unsigned int height, unsigned int bpp,
    			bool bQAReadback, bool bUseFBO, bool bUsePBO) 
	{
        allocateMemory(width, height, bpp, bQAReadback, bUseFBO, bUsePBO);
    }

	virtual ~CheckRender() 
	{
		// Release PBO resources
		glDeleteBuffersARB(1, &m_pboReadback);
		m_pboReadback = 0;            

        free(m_pImageData);
    }

    virtual void 
        allocateMemory( unsigned int width, unsigned int height, unsigned int bpp,
	    	            bool bQAReadback, bool bUseFBO, bool bUsePBO ) 
    {
        m_Width         = width;
        m_Height        = height;
        m_Bpp           = bpp;
        m_bQAReadback   = bQAReadback;
        m_bUseFBO       = bUseFBO;
        m_bUsePBO       = bUsePBO;
        m_PixelFormat   = GL_BGRA;

        // Create the PBO for readbacks
        if (m_bUsePBO) {
            glGenBuffersARB(1, &m_pboReadback);
            glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, m_pboReadback);
            glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, m_Width*m_Height*m_Bpp, NULL, GL_STREAM_READ);
            glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
        }

        m_pImageData = malloc(m_Width*m_Height*m_Bpp);  // This is the image data stored in system memory
    }

    virtual void setExecPath(char *path) {
        strcpy(m_ExecPath, path);
    }
    virtual void EnableQAReadback(bool bStatus) { m_bQAReadback = bStatus; }
    virtual bool IsQAReadback() { return m_bQAReadback; }
    virtual bool IsFBO()        { return m_bUseFBO; }
    virtual bool IsPBO()        { return m_bUsePBO; }

    virtual void setPixelFormat(GLenum format) { m_PixelFormat = format; }
    virtual int  getPixelFormat() { return m_PixelFormat; }
    virtual bool checkStatus(const char *zfile, int line, bool silent) = 0;
    virtual void freeResources() = 0;
    virtual bool readback(GLuint width, GLuint height, void **ppReadBackBuf ) = 0;

    virtual void 
        savePPM(  const char *zfilename, bool bInvert, void **ppReadBuf )
    {
        if (zfilename != NULL) {
            if (bInvert) {
                unsigned char *readBuf;
                unsigned char *writeBuf= (unsigned char *)malloc(m_Width * m_Height * m_Bpp);

                for (unsigned int y=0; y < m_Height; y++) {
                    if (ppReadBuf) {
                        readBuf = *(unsigned char **)ppReadBuf;
                    } else {
                        readBuf = (unsigned char *)m_pImageData;
                    }
                    memcpy(&writeBuf[m_Width*m_Bpp*y], (readBuf+ m_Width*m_Bpp*(m_Height-1-y)), m_Width*m_Bpp);
                }
                // we copy the results back to original system buffer
                if (ppReadBuf) {
                    memcpy(*ppReadBuf, writeBuf, m_Width*m_Height*m_Bpp);
                } else {
                    memcpy(m_pImageData, writeBuf, m_Width*m_Height*m_Bpp);
                }
                free (writeBuf);
            }
            printf("> Saving PPM: <%s>\n", zfilename);
            if (ppReadBuf) {
    		    cutSavePPM4ub(zfilename, *(unsigned char **)ppReadBuf, m_Width, m_Height);
            } else {
    		    cutSavePPM4ub(zfilename, (unsigned char *)m_pImageData, m_Width, m_Height);
            }
        }
    }

    virtual bool 
        PPMvsPPM( const char *src_file, const char *ref_file )
    {
        unsigned char *src_data, *ref_data;
        unsigned long error_count = 0;
        unsigned int width, height;

        char *ref_path = cutFindFilePath(ref_file, m_ExecPath);
        if (src_file == NULL || ref_path == NULL) {
            printf("PPMvsPPM: Aborting comparison\n");
            return false;
        }
        printf("> Comparing (a)rendered: <%s>, (b)reference: <%s>\n", src_file, ref_path);
 
        if (cutLoadPPM4ub(ref_path, &ref_data, &width, &height) != CUTTrue) {
            printf("PPMvsPPM: unable to load ref image file: %s\n", ref_file);
            return false;
        }

        if (cutLoadPPM4ub(src_file, &src_data, &width, &height) != CUTTrue) {
            printf("PPMvsPPM: unable to load src image file: %s\n", src_file);
            return false;
        }

        unsigned long offset;
        for (unsigned int y=0; y < m_Height; y++) {
            for (unsigned int x=0; x < m_Width; x++) {
                int pix_error = 0;
                offset = y*m_Width*m_Bpp + x*m_Bpp;

                if (src_data[offset+0] != ref_data[offset+0]) pix_error++;
                if (src_data[offset+1] != ref_data[offset+1]) pix_error++;
                if (src_data[offset+2] != ref_data[offset+2]) pix_error++;
                if (src_data[offset+3] != ref_data[offset+3]) {} 

                if (pix_error != 0) error_count++;
            }
        }
        if (error_count == 0) { 
            printf("  PASSED!\n"); 
        } else {
            printf("  FAILED! %d errors...\n", (unsigned int)error_count);
        }
        return (error_count == 0);  // returns true if all pixels pass
    }

	virtual void bindFragmentProgram()  = 0;
	virtual void bindRenderPath()       = 0;
	virtual void unbindRenderPath()     = 0;

	virtual void bindTexture()          = 0;
	virtual void unbindTexture()        = 0;

	virtual void bindReadback() {
        if (!m_bQAReadback) {
            printf("CheckRender::bindReadback() uninitialized!\n");
            return;
        }
        if (m_bUsePBO) {
    		glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, m_pboReadback);	// Bind the PBO
        }
	}
	virtual void unbindReadback() {
        if (!m_bQAReadback) {
            printf("CheckRender::unbindReadback() uninitialized!\n");
            return;
        }
        if (m_bUsePBO) {
    		glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, 0);	// Release the bind on the PBO
        }
	}

protected:
    unsigned int m_Width, m_Height, m_Bpp;
    void        *m_pImageData;  // This is the image data stored in system memory
    bool         m_bQAReadback, m_bUseFBO, m_bUsePBO;
    GLuint       m_pboReadback;
    GLenum       m_PixelFormat;
    char         m_ExecPath[256];
};


class CheckBackBuffer : public CheckRender
{
public:
	CheckBackBuffer(unsigned int width, unsigned int height, unsigned int bpp);
	virtual ~CheckBackBuffer();

	virtual bool checkStatus(const char *zfile, int line, bool silent);
    virtual void freeResources() {};
    bool readback( GLuint width, GLuint height, void **ppReadBackBuf );

private:
	virtual void bindFragmentProgram() {}; 
	virtual void bindRenderPath() {};
	virtual void unbindRenderPath() {};

	// bind to the FBO to Texture
	virtual void bindTexture() {}; 

	// release this bind
	virtual void unbindTexture() {}; 
};


// structure defining the properties of a single buffer
struct bufferConfig {
    string name;
    GLenum format;
    int bits;
};

// structures defining properties of an FBO
struct fboConfig {
    string name;
    GLenum colorFormat;
    GLenum depthFormat;
    int redbits;
    int depthBits;
    int depthSamples;
    int coverageSamples;
};

struct fboData {
    GLuint tex; //color texture
    GLuint depthTex; //depth texture
    GLuint fb; // render framebuffer
    GLuint resolveFB; //multisample resolve target
    GLuint colorRB; //color render buffer
    GLuint depthRB; // depth render buffer
};


// CheckFBO - render and verify contents of the FBO
class CheckFBO: public CheckRender
{
public:
	CheckFBO(unsigned int width, unsigned int height, unsigned int bpp);
	CheckFBO(unsigned int width, unsigned int height, unsigned int bpp, fboData &data, fboConfig &config);

	virtual ~CheckFBO();

	virtual bool checkStatus(const char *zfile, int line, bool silent);
	virtual void freeResources();
	virtual bool readback( GLuint width, GLuint height, void **ppReadBackBuf );

	bool initialize(unsigned width, unsigned height, fboConfig & rConfigFBO, fboData & rActiveFBO);
	bool create( GLuint width, GLuint height, fboConfig &config, fboData &data );
	bool createMSAA( GLuint width, GLuint height, fboConfig *p_config, fboData *p_data );
	bool createCSAA( GLuint width, GLuint height, fboConfig *p_config, fboData *p_data );

	// bind the fragment program we'll just render the quad as
	virtual void bindFragmentProgram()
	{
        if (!m_bQAReadback || !m_bUseFBO) {
            printf("CheckFBO::bindFragmentProgram() uninitialized!\n");
            return;
        }
		glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, m_textureProgram);
		glEnable(GL_FRAGMENT_PROGRAM_ARB);
	}

	// bind to the FrameBuffer Object
	virtual void bindRenderPath()
	{
        if (!m_bQAReadback || !m_bUseFBO) {
            printf("CheckFBO::bindReaderPath() uninitialized!\n");
            return;
        }
		glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, m_fboData.fb );
	}

	// release current FrameBuffer Object
	virtual void unbindRenderPath()
	{
        if (!m_bQAReadback || !m_bUseFBO) {
            printf("CheckFBO::unbindReaderPath() uninitialized!\n");
            return;
        }
		glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
	}

	// bind to the FBO to Texture
	virtual void bindTexture()
	{
        if (!m_bQAReadback || !m_bUseFBO) {
            printf("CheckFBO::v() uninitialized!\n");
            return;
        }
		glBindTexture( GL_TEXTURE_2D, m_fboData.tex );
	}

	// release this bind
	virtual void unbindTexture()
	{
        if (!m_bQAReadback || !m_bUseFBO) {
            printf("CheckFBO::unbindTexture() uninitialized!\n");
            return;
        }
		glBindTexture( GL_TEXTURE_2D, 0 );
	}


private:
	fboData		m_fboData;
	fboConfig	m_fboConfig;

	GLuint		m_textureProgram;
	GLuint		m_overlayProgram;
};

#endif // _RENDERCHECK_GL_H_

