/*
* Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/

#include "ImageDX.h"

#include <cuda.h>
#include <cudad3d9.h>

#include <cassert>

const unsigned int ImageDX::aIndexBuffer_[6] = {0, 1, 2, 0, 2, 3};

ImageDX::ImageDX(IDirect3DDevice9 * pDeviceD3D, unsigned int nWidth, unsigned int nHeight, PixelFormat ePixelFormat): pDeviceD3D_(pDeviceD3D)
                                                                                                                    , pTexture_(0)
                                                                                                                    , pSurface_(0)
                                                                                                                    , nWidth_(nWidth)
                                                                                                                    , nHeight_(nHeight)
                                                                                                                    , bIsCudaResource_(false)

{
    assert(0 != pDeviceD3D_);
    
    HRESULT hResult = pDeviceD3D_->CreateTexture(nWidth_, nHeight_, 1, 0, d3dFormat(ePixelFormat), D3DPOOL_DEFAULT, &pTexture_, 0);
    assert(S_OK == hResult);
    assert(0 != pTexture_);
	hResult = pTexture_->GetSurfaceLevel(0, &pSurface_);
    assert(S_OK == hResult);
    assert(0 != pSurface_);

    aVertexBuffer_[0] = VertexStruct(0.0f,           0.0f,            0.0f, 0.0f, 1.0f);
    aVertexBuffer_[1] = VertexStruct((float)nWidth_, 0.0f,            0.0f, 1.0f, 1.0f);
    aVertexBuffer_[2] = VertexStruct((float)nWidth_, (float)nHeight_, 0.0f, 1.0f, 0.0f);
    aVertexBuffer_[3] = VertexStruct(0.0f,           (float)nHeight_, 0.0f, 0.0f, 0.0f);
}

ImageDX::~ImageDX()
{
	pSurface_->Release();
	pTexture_->Release();
}

void
ImageDX::registerAsCudaResource()
{
            // register the Direct3D resources that we'll use
    cuD3D9RegisterResource(pTexture_, 0);
    bIsCudaResource_ = true;

            // we will be write directly to this 2D texture, so we must set the 
            // appropriate flags (to eliminate extra copies during map, but unmap will do copies)
    cuD3D9ResourceSetMapFlags(pTexture_, CU_D3D9_MAPRESOURCE_FLAGS_WRITEDISCARD);
}

void
ImageDX::unregisterAsCudaResource()
{
    cuD3D9UnregisterResource(pTexture_);
	bIsCudaResource_ = false;
}

bool
ImageDX::isCudaResource()
const
{
    return bIsCudaResource_;
}

void
ImageDX::map(CUdeviceptr * ppImageData, unsigned int * pImagePitch)
{
    cuD3D9MapResources(1, reinterpret_cast<IDirect3DResource9 **>(&pTexture_));
	cuD3D9ResourceGetMappedPointer(ppImageData, pTexture_, 0, 0);
	assert(0 != *ppImageData);
	cuD3D9ResourceGetMappedPitch  (pImagePitch, NULL, pTexture_, 0, 0);
	assert(0 != *pImagePitch);
}

void
ImageDX::unmap()
{
    cuD3D9UnmapResources(1, reinterpret_cast<IDirect3DResource9 **>(&pTexture_));
}

void
ImageDX::clear(unsigned char nClearColor)
{
            // Can only be cleared if surface is a CUDA resource
    assert(bIsCudaResource_);
    cuD3D9MapResources(1, (IDirect3DResource9 **)&pTexture_);
    CUdeviceptr  pData = 0;
    unsigned int nSize = 0;
    cuD3D9ResourceGetMappedPointer(&pData, pTexture_, 0, 0);
	assert(0 != pData);
    cuD3D9ResourceGetMappedSize   (&nSize, pTexture_, 0, 0);
	assert(0 != nSize);
	        // clear the surface to solid white
    cuMemsetD8(pData, nClearColor, nSize);
    cuD3D9UnmapResources(1, (IDirect3DResource9 **)&pTexture_);
}

unsigned int
ImageDX::width()
const
{
    return nWidth_;
}

unsigned int
ImageDX::height()
const
{
    return nHeight_;
}

void
ImageDX::render()
const
{
	        // use this index and vertex data throughout
	        // initialize the scene
	pDeviceD3D_->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	pDeviceD3D_->SetRenderState(D3DRS_LIGHTING, FALSE);
	// pDeviceD3D_->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
	pDeviceD3D_->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE );
	pDeviceD3D_->SetFVF(D3DFVF_XYZ|D3DFVF_TEX1|D3DFVF_TEXCOORDSIZE2(0));
	// pDeviceD3D_->SetFVF(D3DFVF_XYZRHW|D3DFVF_TEX1|D3DFVF_TEXCOORDSIZE2(0));
	        // draw the 2d texture
	pDeviceD3D_->SetTexture(0, pTexture_);
	pDeviceD3D_->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_BORDER);
	pDeviceD3D_->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_BORDER);
	pDeviceD3D_->SetSamplerState(0, D3DSAMP_BORDERCOLOR, 0xFFFFFFFF); // set border to solid white
    pDeviceD3D_->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, 4, 2, aIndexBuffer_, D3DFMT_INDEX32, aVertexBuffer_, sizeof(VertexStruct) );
}

D3DFORMAT
ImageDX::d3dFormat(PixelFormat ePixelFormat)
{
    switch (ePixelFormat)
    {
        case LUMINANCE_PIXEL_FORMAT:
            return D3DFMT_L8;
        case BGRA_PIXEL_FORMAT:
            return D3DFMT_A8R8G8B8;
        case UNKNOWN_PIXEL_FORMAT:
            assert(false);
        default:
            assert(false);
    }
    
    assert(false);
    return D3DFMT_UNKNOWN;
}

ImageDX::VertexStruct::VertexStruct(float nX, float nY, float nZ, float nU, float nV)
{
    position[0] = nX;
    position[1] = nY;
    position[2] = nZ;

    texture[0]  = nU;
    texture[1]  = nV;
}

