/*
* Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/

/*
    NV12ToARGB color space conversion CUDA kernel

    This sample uses CUDA to perform a simple NV12 (YUV 4:2:0 planar) 
    source and converts to output in ARGB format
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cutil.h>
#include "cudaProcessFrame.h"

__constant__ uint32 constAlpha;

#define MUL(x,y)    (x*y)
__constant__ float  constHueColorSpaceMat[9];


__device__ void YUV2RGB(uint32 *yuvi, float *red, float *green, float *blue)
{
    float luma, chromaCb, chromaCr;

   // Prepare for hue adjustment
    luma     = (float)yuvi[0];
    chromaCb = (float)((int32)yuvi[1] - 512.0f);
    chromaCr = (float)((int32)yuvi[2] - 512.0f);

   // Convert YUV To RGB with hue adjustment
   *red  = MUL(luma,     constHueColorSpaceMat[0]) + 
           MUL(chromaCb, constHueColorSpaceMat[1]) + 
           MUL(chromaCr, constHueColorSpaceMat[2]);
   *green= MUL(luma,     constHueColorSpaceMat[3]) + 
           MUL(chromaCb, constHueColorSpaceMat[4]) + 
           MUL(chromaCr, constHueColorSpaceMat[5]);
   *blue = MUL(luma,     constHueColorSpaceMat[6]) + 
           MUL(chromaCb, constHueColorSpaceMat[7]) + 
           MUL(chromaCr, constHueColorSpaceMat[8]);
}

__device__ uint32 RGBAPACK(float red, float green, float blue, uint32 alpha)
{
    uint32 ARGBpixel = 0;

   // Clamp final 10 bit results
   red   = min(max(red,   0.0f), 1023.f);
   green = min(max(green, 0.0f), 1023.f);
   blue  = min(max(blue,  0.0f), 1023.f);

    // Convert to 8 bit unsigned integers per color component
    ARGBpixel = (((uint32)blue  >> 2) | 
                (((uint32)green >> 2) << 8)  | 
                (((uint32)red   >> 2) << 16) | (uint32)alpha);

    return  ARGBpixel;
}


// CUDA kernel for outputing the final ARGB output from NV12;
extern "C"
__global__ void NV12ToARGB(uint32 *lumaImage, uint32 nSourcePitch,
                           uint32 *dstARGB,  uint32 nDestPitch,
                           uint32 width, uint32 height)
{
    int32 x, y;
    uint32 yuv101010Pel[2];
    uint32 processingPitch = ((width) + 63) & ~63;
    uint32 dstImagePitch   = nDestPitch >> 2;
    uint8 *lumaImageU8     = (uint8 *)lumaImage;

    processingPitch = nSourcePitch;

    // Pad borders with duplicate pixels, and we multiply by 2 because we process 2 pixels per thread
    x = blockIdx.x * (blockDim.x << 1) + (threadIdx.x << 1);
    y = blockIdx.y *  blockDim.y       +  threadIdx.y;

    if (x >= width)
        x = width - 1;
    if (y >= height)
        y = height - 1;

    // Read 2 Luma components at a time, so we don't waste processing since CbCr are decimated this way.
    // if we move to texture we could read 4 luminance values
    yuv101010Pel[0] = (lumaImageU8[y * processingPitch + x    ]) << 2;
    yuv101010Pel[1] = (lumaImageU8[y * processingPitch + x + 1]) << 2;

    uint32 chromaOffset    = processingPitch * height;
    int32 y_chroma = y >> 1;

    if (y & 1)  // odd scanline ?
    {
        uint32 chromaCb;
        uint32 chromaCr;
        
        chromaCb = lumaImageU8[chromaOffset + y_chroma * processingPitch + x    ];
        chromaCr = lumaImageU8[chromaOffset + y_chroma * processingPitch + x + 1];

        if (y_chroma < ((processingPitch >> 1) - 1)) // interpolate chroma vertically
        {
            chromaCb = (chromaCb + lumaImageU8[chromaOffset + (y_chroma + 1) * processingPitch + x    ] + 1) >> 1;
            chromaCr = (chromaCr + lumaImageU8[chromaOffset + (y_chroma + 1) * processingPitch + x + 1] + 1) >> 1;
        }
        
        yuv101010Pel[0] |= (chromaCb << ( COLOR_COMPONENT_BIT_SIZE       + 2));
        yuv101010Pel[0] |= (chromaCr << ((COLOR_COMPONENT_BIT_SIZE << 1) + 2));

        yuv101010Pel[1] |= (chromaCb << ( COLOR_COMPONENT_BIT_SIZE       + 2));
        yuv101010Pel[1] |= (chromaCr << ((COLOR_COMPONENT_BIT_SIZE << 1) + 2));
    }
    else
    {
        yuv101010Pel[0] |= ((uint32)lumaImageU8[chromaOffset + y_chroma * processingPitch + x    ] << ( COLOR_COMPONENT_BIT_SIZE       + 2));
        yuv101010Pel[0] |= ((uint32)lumaImageU8[chromaOffset + y_chroma * processingPitch + x + 1] << ((COLOR_COMPONENT_BIT_SIZE << 1) + 2));

        yuv101010Pel[1] |= ((uint32)lumaImageU8[chromaOffset + y_chroma * processingPitch + x    ] << ( COLOR_COMPONENT_BIT_SIZE       + 2));
        yuv101010Pel[1] |= ((uint32)lumaImageU8[chromaOffset + y_chroma * processingPitch + x + 1] << ((COLOR_COMPONENT_BIT_SIZE << 1) + 2));
    }
    
    // this steps performs the color conversion
    uint32 yuvi[6];
    float red[2], green[2], blue[2];
   
   yuvi[0] =  (yuv101010Pel[0] &   COLOR_COMPONENT_MASK    );	
   yuvi[1] = ((yuv101010Pel[0] >>  COLOR_COMPONENT_BIT_SIZE)       & COLOR_COMPONENT_MASK); 
   yuvi[2] = ((yuv101010Pel[0] >> (COLOR_COMPONENT_BIT_SIZE << 1)) & COLOR_COMPONENT_MASK);

   yuvi[3] =  (yuv101010Pel[1] &   COLOR_COMPONENT_MASK    );	
   yuvi[4] = ((yuv101010Pel[1] >>  COLOR_COMPONENT_BIT_SIZE)       & COLOR_COMPONENT_MASK); 
   yuvi[5] = ((yuv101010Pel[1] >> (COLOR_COMPONENT_BIT_SIZE << 1)) & COLOR_COMPONENT_MASK);

   // YUV to RGB Transformation conversion
   YUV2RGB(&yuvi[0], &red[0], &green[0], &blue[0]);
   YUV2RGB(&yuvi[3], &red[1], &green[1], &blue[1]);

   // Clamp the results to RGBA
   dstARGB[y * dstImagePitch + x     ] = RGBAPACK(red[0], green[0], blue[0], constAlpha);
   dstARGB[y * dstImagePitch + x + 1 ] = RGBAPACK(red[1], green[1], blue[1], constAlpha);
}


// We call this function to launch the CUDA kernel.  Note for NV12toARGB, the kernel outputs 
// 2 pixels at a time, so that is why the grid size width is half as large as it normally would
extern "C"
CUresult  cudaLaunchNV12toARGB(uint32 *d_srcNV12, uint32 nSourcePitch,
                               uint32 *d_dstARGB, uint32 nDestPitch,
                               uint32 width, uint32 height)
{
    cudaError_t error = cudaSuccess;

    dim3 block(32,16);
    dim3 grid((width+(2*block.x-1))/(2*block.x), (height+(block.y-1))/block.y);

    // Each thread will output (64 bytes of data)
    NV12ToARGB<<<grid, block>>>( reinterpret_cast<uint32 *>(d_srcNV12), nSourcePitch,
                                 d_dstARGB,                            nDestPitch,
                                 width, height );

    error = cudaGetLastError();
    if (error != cudaSuccess) {
        printf("failed! -> NV12ToARGB grid=(%d,%d), block(%d,%d)\n", grid.x, grid.y, block.x, block.y);
    }

    return ((error == cudaSuccess) ? CUDA_SUCCESS : CUDA_ERROR_UNKNOWN);
}

extern "C"
CUresult  updateConstantMemory(float *hueCSC)
{
    cudaError_t error = cudaSuccess;

    // Copy the constants to video memory
    error = cudaMemcpyToSymbol( constHueColorSpaceMat,            
                                &hueCSC[0],
                                9 * sizeof(int32), 0);

    uint32 cudaAlpha      = ((uint32)0xff<< 24);

    error = cudaMemcpyToSymbol( constAlpha,
                                &cudaAlpha,                         
                                sizeof(uint32), 0);

    return ((error == cudaSuccess) ? CUDA_SUCCESS : CUDA_ERROR_UNKNOWN);
}

