/*
* Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/

#include "VideoParser.h"

#include "VideoDecoder.h"
#include "FrameQueue.h"

#include <cstring>
#include <cassert>

VideoParser::VideoParser(VideoDecoder * pVideoDecoder, FrameQueue * pFrameQueue): hParser_(0)
{
    assert(0 != pFrameQueue);
    oParserData_.pFrameQueue   = pFrameQueue;
    assert(0 != pVideoDecoder);
    oParserData_.pVideoDecoder = pVideoDecoder;
    
    CUVIDPARSERPARAMS oVideoParserParameters;  
    memset(&oVideoParserParameters, 0, sizeof(CUVIDPARSERPARAMS));
    oVideoParserParameters.CodecType              = pVideoDecoder->codec();
    oVideoParserParameters.ulMaxNumDecodeSurfaces = pVideoDecoder->maxDecodeSurfaces();
    oVideoParserParameters.pUserData              = &oParserData_;
    oVideoParserParameters.pfnSequenceCallback    = HandleVideoSequence;    // Called before decoding frames and/or whenever there is a format change
    oVideoParserParameters.pfnDecodePicture       = HandlePictureDecode;    // Called when a picture is ready to be decoded (decode order)
    oVideoParserParameters.pfnDisplayPicture      = HandlePictureDisplay;   // Called whenever a picture is ready to be displayed (display order)
    CUresult oResult = cuvidCreateVideoParser(&hParser_, &oVideoParserParameters);
    assert(CUDA_SUCCESS == oResult);
}

int 
CUDAAPI 
VideoParser::HandleVideoSequence(void *pUserData, CUVIDEOFORMAT * pFormat)
{
    VideoParserData * pParserData = reinterpret_cast<VideoParserData *>(pUserData);
    
    if ((pFormat->codec         != pParserData->pVideoDecoder->codec()        ) // codec-type
     || (pFormat->coded_width   != pParserData->pVideoDecoder->frameWidth()   )
     || (pFormat->coded_height  != pParserData->pVideoDecoder->frameHeight()  )
     || (pFormat->chroma_format != pParserData->pVideoDecoder->chromaFormat() ) )
    {
        // We don't deal with dynamic changes in video format
        return 0;
    }
    
    return 1;
}

int 
CUDAAPI 
VideoParser::HandlePictureDecode(void * pUserData, CUVIDPICPARAMS * pPicParams)
{
    VideoParserData * pParserData = reinterpret_cast<VideoParserData *>(pUserData);

    bool bFrameAvailable = pParserData->pFrameQueue->waitUntilFrameAvailable(pPicParams->CurrPicIdx);
    if (!bFrameAvailable)
        return false;
        
    pParserData->pVideoDecoder->decodePicture(pPicParams);
    
    return true;
}

int 
CUDAAPI 
VideoParser::HandlePictureDisplay(void *pUserData, CUVIDPARSERDISPINFO * pPicParams)
{
    // std::cout << *pPicParams << std::endl;
    
    VideoParserData * pParserData = reinterpret_cast<VideoParserData *>(pUserData);
    
    pParserData->pFrameQueue->enqueue(pPicParams);

    return 1;
}

std::ostream &
operator << (std::ostream & rOutputStream, const CUVIDPARSERDISPINFO & rParserDisplayInfo)
{
    rOutputStream << "Picture Index: " << rParserDisplayInfo.picture_index << "\n";
    rOutputStream << "Progressive frame: ";
    if (rParserDisplayInfo.progressive_frame)
        rOutputStream << "true\n";
    else
        rOutputStream << "false\n";
    rOutputStream << "Top field first: ";
    if (rParserDisplayInfo.top_field_first)
        rOutputStream << "true\n";
    else
        rOutputStream << "false\n";
    rOutputStream << "Repeat first field: ";
    if (rParserDisplayInfo.repeat_first_field)
        rOutputStream << "true\n";
    else
        rOutputStream << "false\n";
    rOutputStream << "Time stamp: " << rParserDisplayInfo.timestamp << "\n";

    return rOutputStream;
}

