/*
* Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/

#ifndef NV_VIDEO_PARSER
#define NV_VIDEO_PARSER

#include <nvcuvid.h>

#include <iostream>

class FrameQueue;
class VideoDecoder;

// Wrapper class around the CUDA video-parser API.
//  The video parser consumes a video-data stream and parses it into 
// a) Sequences: Whenever a new sequence or initial sequence header is found
//      in the video stream, the parser calls its sequence-handling callback
//      function.
// b) Decode segments: Whenever a a completed frame or half-frame is found
//      the parser calls its picture decode callback.
// c) Display: Whenever a complete frame was decoded, the parser calls the
//      display picture callback.
//
class VideoParser
{
public:
    // Constructor.
    //
    // Parameters:
    //      pVideoDecoder - pointer to valid VideoDecoder object. This VideoDecoder
    //          is used in the parser-callbacks to decode video-frames.
    //      pFrameQueue - pointer to a valid FrameQueue object. The FrameQueue is used
    //          by  the parser-callbacks to store decoded frames in it.
    VideoParser(VideoDecoder * pVideoDecoder, FrameQueue * pFrameQueue);

private:
    // Struct containing user-data to be passed by parser-callbacks.
    struct VideoParserData
    {
        VideoDecoder * pVideoDecoder;
        FrameQueue   * pFrameQueue;
    };

    // Default constructor. Don't implement.
    explicit
    VideoParser();
    
    // Copy constructor. Don't implement.
    VideoParser(const VideoParser & );
    
    // Assignement operator. Don't implement.
    void
    operator= (const VideoParser & );

    // Called when the decoder encounters a video format change (or initial sequence header)
    // This particular implementation of the callback returns 0 in case the video format changes
    // to something different than the original format. Returning 0 causes a stop of the app.
    static 
    int 
    CUDAAPI 
    HandleVideoSequence(void * pUserData, CUVIDEOFORMAT * pFormat);

    // Called by the video parser to decode a single picture
    // Since the parser will deliver data as fast as it can, we need to make sure that the picture
    // index we're attempting to use for decode is no longer used for display
    static 
    int 
    CUDAAPI 
    HandlePictureDecode(void * pUserData, CUVIDPICPARAMS * pPicParams);

    // Called by the video parser to display a video frame (in the case of field pictures, there may be
    // 2 decode calls per 1 display call, since two fields make up one frame)
    static
    int
    CUDAAPI
    HandlePictureDisplay(void * pUserData, CUVIDPARSERDISPINFO * pPicParams);


    VideoParserData oParserData_;   // instance of the user-data we have passed into the parser-callbacks.
    CUvideoparser   hParser_;       // handle to the CUDA video-parser
    
    friend class VideoSource;
};

std::ostream &
operator << (std::ostream & rOutputStream, const CUVIDPARSERDISPINFO & rParserDisplayInfo);


#endif // NV_VIDEO_PARSER