/*
* Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/

#ifndef NV_VIDEO_SOURCE
#define NV_VIDEO_SOURCE

#include <nvcuvid.h>

#include <string>

// forward declarations
class FrameQueue;
class VideoParser;


// A wrapper class around the CUvideosource entity and API.
//  The CUvideosource manages video-source streams (right now
// via openening a file containing the stream.) After successfully
// opening a video stream, one can query its properties, such as 
// video and audio compression format, frame-rate, etc.
//
// The video-source spawns its own thread for processing the stream.
// The user can register call-back methods for handling chucks of demuxed
// audio and video data.
class VideoSource
{
public:
    // Default constructor. 
    // Parameters:
    //      pFrameQueue - A frame queue object that the decoding
    //          thread and the main render thread use to excange
    //          decoded frames.
    VideoSource(const std::string sFileName, FrameQueue * pFrameQueue);
    
    // Destructor
   ~VideoSource();
   
    CUVIDEOFORMAT
    format()
    const;
   
    // In order to process video-frames, we need to hook up a video-parser
    // object to this source. 
    // Internally we set the CUvideoparser wrapped in rVideoParser on the 
    // user-data struct, that gets passed into the source's callbacks. That 
    // way we can feed the CUDA video parser with the video-data chunks delivered
    // by this source.
    void
    setParser(VideoParser & rVideoParser);

    // Begin processing the video stream.
    void
    start();
    
    // End processing the video stream.
    void
    stop();

    // Has video-processing be started?
    bool
    isStarted();

    // Retrieve dimensions (width, height) of the video source
    void getDimensions(unsigned int &width, unsigned int &height);

private:
    // This struct contains the data we need inside the source's
    // video callback in order to processes the video data.
    struct VideoSourceData
    {
        CUvideoparser hVideoParser;
        FrameQueue  * pFrameQueue;
    };


    // Callback for handling packages of demuxed video data.
    //
    // Parameters:
    //      pUserData - Pointer to user data. We must pass a pointer to a 
    //          VideoSourceData struct here, that contains a valid CUvideoparser
    //          and FrameQueue.
    //      pPacket - video-source data packet. 
    //
    // NOTE: called from a different thread that doesn't not have a cuda context
    //
    static
    int 
    CUDAAPI
    HandleVideoData(void * pUserData, CUVIDSOURCEDATAPACKET * pPacket);

    // Default constructor. Don't implemenent.
    VideoSource();
    
    // Copy constructor. Don't implement.
    VideoSource(const VideoSource & );
    
    // Assignment operator. Don't implement.
    void
    operator= (const VideoSource & );
    
    VideoSourceData oSourceData_;       // Instance of the user-data struct we use in the video-data handle callback.
    CUvideosource   hVideoSource_;      // Handle to the CUDA video-source object.
};

std::ostream &
operator << (std::ostream & rOutputStream, const CUVIDEOFORMAT & rCudaVideoFormat);

#endif // NV_VIDEO_SOURCE