/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

 /* CUmoduleManager manages loading CUBIN, initializing CUBIN kernel function,
  * initializating CUDA kernel function pointers, and obtaining global memory 
  * addresses (i.e. constants).
  */

#include <stdio.h>
#include <string.h>

#include <cuda.h>
#include <builtin_types.h>

#include <cutil.h>

#include "cudaModuleMgr.h"


CUmoduleManager::CUmoduleManager(const char *filename_cubin, 
                                 const char *exec_path, 
                                 int nKernels, 
                                 int nGlobalMem) 
    : nMaxKernels_(nKernels), nMaxGlobalMem_(nGlobalMem)
{
    CUfunction cuFunction = 0;
    CUresult   cuStatus = CUDA_SUCCESS;
	int kernel_count = 0;

    char* module_path;

    module_path = cutFindFilePath(filename_cubin, exec_path);
    if (module_path == 0) {
        printf("CUmoduleManager CUBIN <%s> not found!\n", filename_cubin);
        cutFree(module_path);
        return;
    }

    cuStatus = cuModuleLoad(&cuModule_, module_path);
    cutFree(module_path);

    printf("CUmoduleManager() - <%s> initialized\t= %s", 
            filename_cubin, 
            (cuStatus == CUDA_SUCCESS) ? "SUCCESS!\n" : "FAILED!\n");

    if (cuStatus == CUDA_SUCCESS) {
        pCudaKernels_  = new CudaKernels  [nMaxKernels_];
        pGlobalMem_    = new CudaGlobalMem[nMaxGlobalMem_];
        nLastKernel_   = 0;
        nLastGlobalMem_ = 0;
    }
}

CUmoduleManager::~CUmoduleManager()
{
    delete pCudaKernels_;
    delete pGlobalMem_;
}

CUresult
CUmoduleManager::GetCudaFunction(const char *func_name, CUfunction *fpCudaKernel)
{
    int idx = nLastKernel_;

    CUresult cuStatus = cuModuleGetFunction( &(pCudaKernels_[idx].fpCuda), cuModule_, func_name );
    printf("CUmoduleManager::GetCudaFunction<%25s>\t= 0x%08x\n", func_name, (unsigned long long)(pCudaKernels_[idx].fpCuda));
    pCudaKernels_[idx].func_name = func_name;

    if (fpCudaKernel) 
        *fpCudaKernel = pCudaKernels_[idx].fpCuda;

    nLastKernel_++;

    return cuStatus;
}

CUresult
CUmoduleManager::GetCudaDevicePtr(const char *address_name, CUdeviceptr *pGlobalMem)
{
    int idx = nLastGlobalMem_;

    CUresult cuStatus = cuModuleGetGlobal( &(pGlobalMem_[idx].devicePtr),  &(pGlobalMem_[idx].nBytes), cuModule_, address_name);
    printf("CUmoduleManager::GetCudaDevicePtr<%25s>\t= 0x%08x\n", address_name, (unsigned long long)(pGlobalMem_[idx].devicePtr) );
    pGlobalMem_[idx].address_name = address_name;

    if (pGlobalMem) 
        *pGlobalMem = pGlobalMem_[idx].devicePtr;

    nLastGlobalMem_++;

    return cuStatus;
}

int CUmoduleManager::findIndex_CudaKernels(const char *func_name)
{
    int found = -1;

    for (int i=0; i < nLastKernel_; i++) {
        if (pCudaKernels_[i].func_name.compare(func_name)) {
            found = i;
            break;
        }
    }

    return found;
}


int CUmoduleManager::findIndex_GlobalMem(const char *address_name)
{
    int found = -1;

    for (int i=0; i < nLastGlobalMem_; i++) {
        if (pGlobalMem_[i].address_name.compare(address_name)) {
            found = i;
            break;
        }
    }

    return found;
}

struct kernel_parameters {

};


// TODO, perhaps we should set this up such that we use templates to pass in a 
// structure to the function parameter
CUresult CUmoduleManager::launchKernel(CUfunction fpFunc, dim3 block, dim3 grid)
{
    CUresult error = CUDA_SUCCESS;
/*
    dim3 block(16,16);
    dim3 grid((width+63)/(2*block.x), (height+15)/block.y); 

    // setup execution parameters
    CU_SAFE_CALL(cuFuncSetBlockShape( fpFunc, block.x, block.y, 1 ));
    int offset = 0;
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 0,  d_srcNV12 ));     offset += sizeof(d_srcNV12);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 4,  nSourcePitch ));  offset += sizeof(nSourcePitch);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 8,  d_dstARGB ));     offset += sizeof(d_dstARGB);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 12, nDestPitch ));    offset += sizeof(nDestPitch);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 16, width ));         offset += sizeof(width);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 20, height ));        offset += sizeof(height);
    CU_SAFE_CALL(cuParamSetSize     ( fpFunc,     offset));           

    error = cuLaunchGrid( fpFunc, grid.x, grid.y );
*/

    return error;
}
