/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

 /* This example demonstrates how to use the Video Decode Library with CUDA 
  * bindings to interop between CUDA and DX9 textures for the purpose of post
  * processing video.
  */

#include "cudaProcessFrame.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef _WIN32
#define WINDOWS_LEAN_AND_MEAN
#include <windows.h>
#endif
#include <d3dx9.h>

#include <cuda.h>
#include <builtin_types.h>
#include <cuda_runtime_api.h>

#include <cudad3d9.h>
#include <cuda_d3d9_interop.h>
#include <cutil.h>

// These store the matrix for YUV2RGB transformation
__constant__ float  constHueColorSpaceMat[9];
__constant__ float  constAlpha;


// Loads the CUBIN module and grabs the function pointers
extern "C" 
CUresult InitCudaModule(char *filename_cubin, char *exec_path, CUmodule *pModule)
{
    CUfunction cuFunction = 0;
    CUresult   cuStatus = CUDA_SUCCESS;
	int kernel_count = 0;

    char* module_path;

    module_path = cutFindFilePath(filename_cubin, exec_path);
    if (module_path == 0) {
        cuStatus = CUDA_ERROR_NOT_FOUND;
        return CUDA_ERROR_INVALID_VALUE;
    }

    cuStatus = cuModuleLoad(pModule, module_path);
    cutFree(module_path);
    if ( CUDA_SUCCESS != cuStatus ) {
        return CUDA_ERROR_FILE_NOT_FOUND;
    }
    printf("InitCudaModule   - <%25s>\t= %s", filename_cubin, (cuStatus == CUDA_SUCCESS) ? "SUCCESS!\n" : "FAILED!\n");

    return cuStatus;
}

extern "C"
CUresult InitCudaFunction(char *func_name, CUmodule *pModule, CUfunction *pCudaFunction)
{
    CUresult cuStatus = cuModuleGetFunction( pCudaFunction,      *pModule, func_name );
    printf("InitCudaFunction - <%25s>\t= 0x%08x\n", func_name, (unsigned long long)(*pCudaFunction));

    return cuStatus;
}


extern "C"
CUresult  updateConstantMemory_drvapi(CUmodule module, float *hueCSC)
{
    CUdeviceptr d_constHueCSC, d_constAlpha;
    unsigned int d_cscBytes, d_alphaBytes;

    // First grab the global device pointers from the CUBIN
    cuModuleGetGlobal(&d_constHueCSC,  &d_cscBytes  , module, "constHueColorSpaceMat");
    cuModuleGetGlobal(&d_constAlpha ,  &d_alphaBytes, module, "constAlpha"           );

    CUresult error = CUDA_SUCCESS;

    // Copy the constants to video memory
    error = cuMemcpyHtoD( d_constHueCSC,            
                          reinterpret_cast<const void *>(hueCSC),
                          d_cscBytes);

    uint32 cudaAlpha      = ((uint32)0xff<< 24);

    error = cuMemcpyHtoD( constAlpha,
                          reinterpret_cast<const void *>(&cudaAlpha),                         
                          d_alphaBytes);

    return error;
}

extern "C"
void setColorSpaceMatrix(eColorSpace CSC, float *hueCSC, float hue)
{
    float hueSin = sin(hue);
    float hueCos = cos(hue);

    if (CSC == ITU601) {
        //CCIR 601
	    hueCSC[0] = 1.1644f;
	    hueCSC[1] = hueSin * 1.5960f;
	    hueCSC[2] = hueCos * 1.5960f;
	    hueCSC[3] = 1.1644f;
	    hueCSC[4] = (hueCos * -0.3918f) - (hueSin * 0.8130f);
	    hueCSC[5] = (hueSin *  0.3918f) - (hueCos * 0.8130f);  
	    hueCSC[6] = 1.1644f;
	    hueCSC[7] = hueCos *  2.0172f;
	    hueCSC[8] = hueSin * -2.0172f;
    } else if (CSC == ITU709) {
        //CCIR 709
        hueCSC[0] = 1.0f;
        hueCSC[1] = hueSin * 1.57480f;
        hueCSC[2] = hueCos * 1.57480f;
        hueCSC[3] = 1.0;
        hueCSC[4] = (hueCos * -0.18732f) - (hueSin * 0.46812f);
        hueCSC[5] = (hueSin *  0.18732f) - (hueCos * 0.46812f);  
        hueCSC[6] = 1.0f;
        hueCSC[7] = hueCos *  1.85560f;
        hueCSC[8] = hueSin * -1.85560f;    
    }
}

// We call this function to launch the CUDA kernel.  Note for NV12toARGB, the kernel outputs 
// 2 pixels at a time, so that is why the grid size width is half as large as it normally would
extern "C"
CUresult  cudaLaunchNV12toARGBDrv(  CUdeviceptr d_srcNV12, uint32 nSourcePitch,
                                    CUdeviceptr d_dstARGB, uint32 nDestPitch,
                                    uint32 width, uint32 height,
                                    CUfunction fpFunc)
{
    dim3 block(32,16);
    dim3 grid((width+(2*block.x-1))/(2*block.x), (height+(block.y-1))/block.y); 

    // setup execution parameters
    CUresult error = CUDA_SUCCESS;
    CU_SAFE_CALL(cuFuncSetBlockShape( fpFunc, block.x, block.y, 1 ));
    int offset = 0;
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 0,  d_srcNV12 ));     offset += sizeof(d_srcNV12);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 4,  nSourcePitch ));  offset += sizeof(nSourcePitch);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 8,  d_dstARGB ));     offset += sizeof(d_dstARGB);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 12, nDestPitch ));    offset += sizeof(nDestPitch);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 16, width ));         offset += sizeof(width);
    CU_SAFE_CALL(cuParamSeti        ( fpFunc, 20, height ));        offset += sizeof(height);
    CU_SAFE_CALL(cuParamSetSize     ( fpFunc,     offset));           

    error = cuLaunchGrid( fpFunc, grid.x, grid.y );

    return error;
}
