/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

 /* This example demonstrates how to use the Video Decode Library with CUDA 
  * bindings to interop between CUDA and DX9 textures for the purpose of post
  * processing video.
  */

#ifdef _WIN32
#define WINDOWS_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#endif
#include <d3dx9.h>

#include <math.h>
#include <cuda.h>
#include <builtin_types.h>

#include <cudad3d9.h>
#include <cutil.h>
#include <cuda_runtime_api.h>

#include "FrameQueue.h"
#include "VideoSource.h"
#include "VideoParser.h"
#include "VideoDecoder.h"
#include "ImageDX.h"

#include "cudaProcessFrame.h"
#include "cudaModuleMgr.h"

#include <memory>
#include <iostream>
#include <cassert>

#define VIDEO_SOURCE_FILE "plush1_720p_10s.m2v"

#define CU_CTX_BLOCKING_SYNC 4
#define USE_DRVAPI           1

bool                gbDone    = false;
bool                gbRunning = false;

IDirect3D9		 * gpD3D; // Used to create the D3DDevice
IDirect3DDevice9 * gpD3DDevice;

// These are CUDA function pointers to the CUDA kernels
CUmoduleManager  * gpCudaModule;

CUmodule           cuModNV12toARGB  = 0;
CUfunction         gfpNV12toARGB    = 0;
CUfunction         gfpNV12Passthru  = 0;

CUcontext          goContext = 0;
CUdevice           goDevice  = 0;

eColorSpace        geColorSpace = ITU601;
float              gnHue        = 0.0f;

FrameQueue   * gpFrameQueue   = 0;
VideoSource  * gpVideoSource  = 0;
VideoParser  * gpVideoParser  = 0;
VideoDecoder * gpVideoDecoder = 0;

ImageDX      * gpImageDX      = 0;


unsigned int gnWindowWidth  = 0;
unsigned int gnWindowHeight = 0;

unsigned int gnClientAreaWidth  = 0;
unsigned int gnClientAreaHeight = 0;

unsigned int gnVideoWidth  = 0;
unsigned int gnVideoHeight = 0;

unsigned int g_FrameCount = 0;
unsigned int g_fpsCount = 0;      // FPS count for averaging
unsigned int g_fpsLimit = 16;      // FPS limit for sampling
unsigned int timer;

// Forward declarations
void initializeD3D( HWND hWnd );
void initializeTexture(unsigned int nWidth, unsigned int nHeight);

void loadVideoSource(const char * video_file, unsigned int & width, unsigned int & height);
void initializeCudaVideo();

void terminateCudaVideo();

void copyDecodedFrameToTexture();
void cudaPostProcessFrame(CUdeviceptr *ppDecodedFrame, unsigned int nDecodedPitch, 
                          CUdeviceptr *ppTextureData,  unsigned int nTexturePitch,
                          CUmodule cuModNV12toARGB, CUfunction fpCudaKernel);
void drawScene();
void cleanup();
void renderVideoFrame( HWND hWnd );

LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);


void computeFPS( HWND hWnd )
{
	CUT_SAFE_CALL(cutStopTimer(timer));  

    if (gbRunning) {
        g_fpsCount++;
        if (!gpFrameQueue->isEndOfDecode())
            g_FrameCount++;
    }

    char sFPS[256];
    std::string sDecodeStatus;

    if (gpFrameQueue->isEndOfDecode()) {
        sDecodeStatus = "STOP (End of File)\0";
    } else {
        if (!gbRunning) {
            sDecodeStatus = "PAUSE\0";
            sprintf(sFPS, "CUDA/D3D9 Video Decoder [%s] - [Frame %d]",
                    sDecodeStatus.c_str(), g_FrameCount);
            SetWindowText(hWnd, sFPS);
            UpdateWindow(hWnd);
        } else {
            sDecodeStatus = "PLAY\0";
        }
    }

    if (g_fpsCount == g_fpsLimit) {
        float ifps = 1.f / (cutGetAverageTimerValue(timer) / 1000.f);

        sprintf(sFPS, "CUDA/D3D9 Video Decoder [%s] - [%3.1f fps, Frame %d]",
                sDecodeStatus.c_str(), ifps, g_FrameCount);
        SetWindowText(hWnd, sFPS);

        CUT_SAFE_CALL(cutResetTimer(timer));  
        g_fpsCount = 0;
    }

	CUT_SAFE_CALL(cutStartTimer(timer));  
}

bool checkComputeCapable(CUdevice dev)
{
    cudaDeviceProp deviceProp;

    CUDA_SAFE_CALL(cudaGetDeviceProperties(&deviceProp, dev));

	if(deviceProp.major > 1 || deviceProp.minor > 0)
    {
		printf("cudaVideoDecode:\n  <%s> CUDA %d.%d device %d detected\n", 
				deviceProp.name, deviceProp.major, deviceProp.minor, dev);
		return true;
	} else {
		printf("cudaVideoDecode:\n  <%s> CUDA %d.%d device %d detected\n", 
				deviceProp.name, deviceProp.major, deviceProp.minor, dev);
		printf("  This device not CUDA 1.1 capable, exiting\n\n");
		printf("  Test PASSED\n");
		return false;
	}
}



int main(int argc, char* argv[])
{
    CUT_SAFE_CALL(cutCreateTimer(&timer));
    CUT_SAFE_CALL(cutResetTimer(timer));  

    char* video_file;
    video_file = cutFindFilePath(VIDEO_SOURCE_FILE, argv[0]);
    if (video_file == 0) {
        printf("cudaVideoDecode: unable to find file: <%s>, aborting\n", VIDEO_SOURCE_FILE);
        CUT_EXIT(argc, argv);
    }

    // default video file loaded by this sample
    std::string sFileName( video_file );
            // if there's a command-line argument
            // use it as filename.
    if (argc > 1)
        sFileName = argv[1];
	        // create window (after we know the size of the input file size)
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "CUDA/D3D9 Video Decoder", NULL };
    RegisterClassEx( &wc );
            // Find out the video size 
    loadVideoSource(sFileName.c_str(), gnVideoWidth  , gnVideoHeight );
            // figure out the window size we must create to get a *client* area 
            // that is of the size requested by m_dimensions.
    RECT adjustedWindowSize;
    DWORD dwWindowStyle = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    SetRect(&adjustedWindowSize, 0, 0, gnVideoWidth  , gnVideoHeight);
    AdjustWindowRect(&adjustedWindowSize, dwWindowStyle, false);

    gnWindowWidth  = adjustedWindowSize.right  - adjustedWindowSize.left;
    gnWindowHeight = adjustedWindowSize.bottom - adjustedWindowSize.top;
            // Create the application's window
    HWND hWnd = CreateWindow( wc.lpszClassName, "CUDA/D3D9 Video Decoder",
                              dwWindowStyle,
                              0, 0, 
                              gnWindowWidth, 
                              gnWindowHeight,
                              NULL, NULL, wc.hInstance, NULL );
            // Initialize CUDA
    cuInit(0);
            // Initialize Direct3D
    initializeD3D(hWnd);
            // Create CUDA Device
    cuD3D9CtxCreate(&goContext, &goDevice, CU_CTX_BLOCKING_SYNC, gpD3DDevice);

			// Check for Compute 1.1 capability before running
	if (!checkComputeCapable(goDevice)) {
				// Unregister windows class
		UnregisterClass( wc.lpszClassName, wc.hInstance );
        CUT_EXIT(argc, argv);
	}

                // Initialize CUDA releated Driver API
    gpCudaModule = new CUmoduleManager("NV12ToARGB_drvapi.cubin", argv[0], 2, 2);

    gpCudaModule->GetCudaFunction("NV12ToARGB_drvapi",   &gfpNV12toARGB);
    gpCudaModule->GetCudaFunction("NV12Passthru_drvapi", &gfpNV12Passthru);

    initializeCudaVideo();
    
    initializeTexture(gpVideoDecoder->targetWidth(), gpVideoDecoder->targetHeight());

    gpVideoSource->start();
    gbRunning = true;

    ShowWindow(hWnd, SW_SHOWDEFAULT);
    UpdateWindow(hWnd);
	        // the main loop

	CUT_SAFE_CALL(cutStartTimer(timer));  

    while(!gbDone) 
	{
        MSG msg;
        ZeroMemory( &msg, sizeof(msg) );
        while( msg.message!=WM_QUIT )
        {
            if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
            {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
            }
            else
			{
                renderVideoFrame( hWnd );
			}
        }
    } // while loop
    gpFrameQueue->endDecode();
    gpVideoSource->stop();
			// clean up CUDA and D3D resources
    terminateCudaVideo();
	cleanup();
			// Unregister windows class
	UnregisterClass( wc.lpszClassName, wc.hInstance );

    // If we are doing the QAtest, we quite without prompting
    if( cutCheckCmdLineFlag( argc, (const char**) argv, "qatest") ||
        cutCheckCmdLineFlag( argc, (const char**) argv, "noprompt"))
    {
        exit(0);
    }

	return 0;
}

// Initialize Direct3D
void
initializeD3D(HWND hWnd) 
{
			// Create the D3D object.
    gpD3D = Direct3DCreate9(D3D_SDK_VERSION);
    assert(gpD3D != 0);

    // Get primary display identifier
    D3DADAPTER_IDENTIFIER9 adapterId;
    
    // Find the first CUDA capable device
    unsigned int adapter;
    for(adapter = 0; adapter < gpD3D->GetAdapterCount(); adapter++)
    {
        D3DADAPTER_IDENTIFIER9 ident;
        int device;
        gpD3D->GetAdapterIdentifier(adapter, 0, &adapterId);
        cuD3D9GetDevice(&device, ident.DeviceName);
    }
            // require that we found a cuda-compatible device to work on
    assert(adapter == gpD3D->GetAdapterCount());
			// Create the D3D Display Device
    RECT                  rc;       
    GetClientRect(hWnd, &rc);
    D3DDISPLAYMODE        d3ddm;    
    gpD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3ddm);
    D3DPRESENT_PARAMETERS d3dpp;    
    ZeroMemory(&d3dpp, sizeof(d3dpp));

    d3dpp.Windowed               = true; // fullscreen or windowed?
    d3dpp.BackBufferCount        = 1;
    d3dpp.SwapEffect             = D3DSWAPEFFECT_DISCARD;
    d3dpp.hDeviceWindow          = hWnd;
    d3dpp.BackBufferWidth	     = rc.right  - rc.left;
    d3dpp.BackBufferHeight       = rc.bottom - rc.top;
    d3dpp.BackBufferFormat       = d3ddm.Format;
    d3dpp.FullScreen_RefreshRateInHz = 0; // set to 60 for fullscreen, and also don't forget to set Windowed to FALSE
    d3dpp.PresentationInterval   = D3DPRESENT_INTERVAL_ONE; // D3DPRESENT_DONOTWAIT;
    
    HRESULT eResult = gpD3D->CreateDevice(0, D3DDEVTYPE_HAL, hWnd,
                                           D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_MULTITHREADED,
                                           &d3dpp, &gpD3DDevice);
	assert(S_OK == eResult);
}


// If we want to test to use only the LUMA channel only, uncoment this out, and comment 
// following lines below
//#define TEXTURE_FORMAT ImageDX::LUMINANCE_PIXEL_FORMAT
//#define TEXTURE_BPP 1

#define TEXTURE_FORMAT ImageDX::BGRA_PIXEL_FORMAT
#define TEXTURE_BPP 4


// Initializes Direct3D Textures (allocation and initialization)
void
initializeTexture(unsigned int nWidth, unsigned int nHeight)
{
    gpImageDX = new ImageDX(gpD3DDevice, nWidth, nHeight, TEXTURE_FORMAT);
	gpImageDX->registerAsCudaResource();
	gpImageDX->clear();
}

void
loadVideoSource(const char *video_file, unsigned int &width, unsigned int &height)
{
    std::auto_ptr<FrameQueue> apFrameQueue(new FrameQueue);
    std::auto_ptr<VideoSource> apVideoSource(new VideoSource(video_file, apFrameQueue.get()));

            // retrieve the video source (width,height)    
    apVideoSource->getDimensions(width, height);

    std::cout << apVideoSource->format() << std::endl;

    gpFrameQueue  = apFrameQueue.release();
    gpVideoSource = apVideoSource.release();
}

void
initializeCudaVideo()
{
    std::auto_ptr<VideoDecoder> apVideoDecoder(new VideoDecoder(gpVideoSource->format()));
    std::auto_ptr<VideoParser> apVideoParser(new VideoParser(apVideoDecoder.get(), gpFrameQueue));
    gpVideoSource->setParser(*apVideoParser.get());
    
    gpVideoParser  = apVideoParser.release();
    gpVideoDecoder = apVideoDecoder.release();
}


void 
terminateCudaVideo()
{
    delete gpVideoParser;
    delete gpVideoDecoder;
    delete gpVideoSource;
    delete gpFrameQueue;
}

		// Run the Cuda part of the computation
void copyDecodedFrameToTexture(unsigned int &nRepeats)
{
    CUVIDPARSERDISPINFO oDisplayInfo;
    if (gpFrameQueue->dequeue(&oDisplayInfo))
    {
        nRepeats = oDisplayInfo.repeat_first_field;
        CUVIDPROCPARAMS oVideoProcessingParameters;
        memset(&oVideoProcessingParameters, 0, sizeof(CUVIDPROCPARAMS));
        
        oVideoProcessingParameters.progressive_frame = oDisplayInfo.progressive_frame;
        oVideoProcessingParameters.top_field_first   = oDisplayInfo.top_field_first;
        oVideoProcessingParameters.second_field = 0;
       
        CUdeviceptr	 pDecodedFrame = 0;
        unsigned int nDecodedPitch = 0;
        unsigned int nWidth = 0;
        unsigned int nHeight = 0;
        
				// map decoded video frame to CUDA surfae
        gpVideoDecoder->mapFrame(oDisplayInfo.picture_index, &pDecodedFrame, &nDecodedPitch, &oVideoProcessingParameters);
        nWidth  = gpVideoDecoder->targetWidth();
        nHeight = gpVideoDecoder->targetHeight();
        assert(gpImageDX->width()  == nWidth);
        assert(gpImageDX->height() == nHeight);
				// map DirectX texture to CUDA surface
		CUdeviceptr  pTextureData  = 0;
		unsigned int nTexturePitch = 0;
        gpImageDX->map(&pTextureData, &nTexturePitch);

#if 1
        // perform post processing on the CUDA surface (performs colors space conversion and post processing)
        // comment this out if we inclue the line of code seen above 
		cudaPostProcessFrame(&pDecodedFrame, nDecodedPitch, &pTextureData, nTexturePitch, gpCudaModule->cuModule_, gfpNV12toARGB);
#else
	    // copy decoded frame surface into texture surface (uncomment this out if we only want to display Luminance only)
	    cuMemcpyDtoD(pTextureData, pDecodedFrame, nDecodedPitch * nHeight);
#endif

			// unmap the texture surface
		gpImageDX->unmap();
                // unmap video frame
		gpVideoDecoder->unmapFrame(pDecodedFrame);
                // mark frame for re-use in decoder
        gpFrameQueue->releaseFrame(&oDisplayInfo);
    } 
            // check if decoding has come to an end.
            // if yes, signal the app to shut down.
    if (!gpVideoSource->isStarted() || gpFrameQueue->isEndOfDecode())
    {
        // Let's just stop, and allow the user to quit, so they can at least see the results
        gpVideoSource->stop();
	}
}

// This is the CUDA stage for Video Post Processing.  Last stage takes care of the NV12 to ARGB
void
cudaPostProcessFrame(CUdeviceptr * ppDecodedFrame, unsigned int nDecodedPitch, 
                     CUdeviceptr * ppTextureData,  unsigned int nTexturePitch,
                     CUmodule cuModNV12toARGB, CUfunction fpCudaKernel)
{
            // CCIR 601/709
    float hueColorSpaceMat[9];

    uint32 nWidth  = gpVideoDecoder->targetWidth();
    uint32 nHeight = gpVideoDecoder->targetHeight();

    // TODO: Stage for handling video post processing

            // Final Stage: NV12toARGB color space conversion
    setColorSpaceMatrix (geColorSpace,    hueColorSpaceMat, gnHue);
	
#if USE_DRVAPI
    CUresult eResult;
    updateConstantMemory_drvapi(cuModNV12toARGB, hueColorSpaceMat       );
    eResult = cudaLaunchNV12toARGBDrv  (*ppDecodedFrame, nDecodedPitch,
                                        *ppTextureData,  nTexturePitch,
                                        nWidth, nHeight, fpCudaKernel);
#else
    CUresult eResult;
    updateConstantMemory( hueColorSpaceMat );
    eResult = cudaLaunchNV12toARGB     (reinterpret_cast<uint32 *>(static_cast<unsigned long long>(*ppDecodedFrame)), nDecodedPitch,
                                        reinterpret_cast<uint32 *>(static_cast<unsigned long long>(*ppTextureData)),  nTexturePitch,
                                        nWidth, nHeight);
    assert(CUDA_SUCCESS == eResult);
#endif
}

// Draw the final result on the screen
void drawScene()
{
	        // initialize the scene
	gpD3DDevice->BeginScene();
	gpD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET, 0, 1.0f, 0);	
	        // render image
    gpImageDX->render();
	        // end the scene
	gpD3DDevice->EndScene();
	gpD3DDevice->Present(NULL, NULL, NULL, NULL);
}

// Release all previously initialized objects
void cleanup()
{
	gpImageDX->unregisterAsCudaResource();
	delete gpImageDX;
			// destroy the D3D device
	gpD3DDevice->Release();
}

// Launches the CUDA kernels to fill in the texture data
void renderVideoFrame( HWND hWnd )
{
    static unsigned int nRepeatFrame = 0;

    if (nRepeatFrame > 0)
    {
        drawScene();
        --nRepeatFrame;
    }
    if (0 != gpFrameQueue)
    {
                // if not running, we simply don't copy
                // new frames from the decoder 
        if (gbRunning)
	        copyDecodedFrameToTexture(nRepeatFrame);
		    	// draw the scene using them
	    drawScene();
	}

    // Pass the Windows handle to show Frame Rate on the window title
    computeFPS( hWnd );
}

// The window's message handler
static LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg)
    {
        case WM_KEYDOWN:
            switch (wParam) 
            {
                        // use ESC to quit application
                case VK_ESCAPE: 
			    {
				    gbDone = true;
	                PostQuitMessage(0);
				    return 0;
			    }
                break;
                        // use space to pause playback
                case VK_SPACE:
                {
                    gbRunning = !gbRunning;
                }
                break;
            }
            break;
                // resize the window
                // even though we disable resizing (see next event handler below)
                // we need to be able to handle this event properly because the
                // windowing system calls it.
        case WM_SIZE:
        {
                    // store new window size in our globals.
            gnClientAreaWidth  = GET_X_LPARAM(lParam);
            gnClientAreaHeight = GET_Y_LPARAM(lParam);

            D3DVIEWPORT9 oViewport;
            oViewport.X = 0;
            oViewport.Y = 0;
            oViewport.Width  = gnClientAreaWidth;
            oViewport.Height = gnClientAreaHeight;
            oViewport.MaxZ = 1.0f;
            oViewport.MinZ = 0.0f;
            
            gpD3DDevice->SetViewport(&oViewport);
            
            D3DMATRIX oViewMatrix = {1.0f, 0.0f, 0.0f, 0.0f, 
                                     0.0f, 1.0f, 0.0f, 0.0f, 
                                     0.0f, 0.0f, 1.0f, 0.0f, 
                                     0.0f, 0.0f, 0.0f, 1.0f};
                    // scale viewport to be of window width and height
            oViewMatrix._11 = 2.0f/gnClientAreaWidth; 
            oViewMatrix._22 = 2.0f/gnClientAreaHeight;
                    // translate viewport so that lower left corner represents
                    // (0, 0) and upper right corner is (width, height)
            oViewMatrix._41 = -1.0f - 1.0f / gnClientAreaWidth; 
            oViewMatrix._42 = -1.0f + 1.0f / gnClientAreaHeight;
            if (0 != gpD3DDevice)
                gpD3DDevice->SetTransform(D3DTS_VIEW, &oViewMatrix);
            
            renderVideoFrame( hWnd );
            
            return 0;   // Jump Back
        }
                // disallow resizing.
        case WM_GETMINMAXINFO:
        {
            MINMAXINFO * pMinMaxInfo = reinterpret_cast<MINMAXINFO *>(lParam);
            
            pMinMaxInfo->ptMinTrackSize.x = gnWindowWidth;
            pMinMaxInfo->ptMinTrackSize.y = gnWindowHeight;
            
            pMinMaxInfo->ptMaxTrackSize.x = gnWindowWidth;
            pMinMaxInfo->ptMaxTrackSize.y = gnWindowHeight;
        
            return 0;
        }
        case WM_DESTROY:
			gbDone = true;
            PostQuitMessage(0);
            return 0;
        case WM_PAINT:
            ValidateRect(hWnd, NULL);
            return 0;
    }
    return DefWindowProc(hWnd, msg, wParam, lParam);
}

